import json
from os.path import exists
from typing import List

from sonusai import SonusAIError
from sonusai import logger


def load_mixid(name: str, mixdb: dict) -> List[int]:
    if not name:
        mixid = list(range(len(mixdb['mixtures'])))
    else:
        if not exists(name):
            logger.exception(f'{name} does not exist')
            raise SonusAIError

        with open(name, encoding='utf-8') as f:
            mixid = json.load(f)
            if not isinstance(mixid, dict) or 'mixid' not in mixid.keys():
                logger.exception(f'Could not find ''mixid'' in {name}')
                raise SonusAIError
            mixid = mixid['mixid']

    return mixid



from os.path import expandvars

import numpy as np
import sox

from sonusai import SonusAIError
from sonusai import logger
from sonusai import mixture


def read_audio(name: str, dither: bool = True) -> np.ndarray:
    expanded_name = expandvars(name)

    try:
        # Read in and convert to desired format
        inp = sox.Transformer()
        inp.set_output_format(rate=mixture.sample_rate, bits=mixture.bit_depth, channels=mixture.channel_count)
        inp.set_globals(dither=dither)
        return inp.build_array(input_filepath=expanded_name,
                               sample_rate_in=int(sox.file_info.sample_rate(expanded_name)))

    except Exception as e:
        if name != expanded_name:
            logger.exception(f'Error reading {name} (expanded: {expanded_name}): {e}')
        else:
            logger.exception(f'Error reading {name}: {e}')
        raise SonusAIError


def apply_augmentation(audio_in: np.ndarray,
                       augmentation: dict,
                       length_common_denominator: int,
                       dither: bool = True) -> np.ndarray:
    try:
        # Apply augmentations
        tfm = sox.Transformer()
        tfm.set_input_format(rate=mixture.sample_rate, bits=mixture.bit_depth, channels=mixture.channel_count)
        tfm.set_output_format(rate=mixture.sample_rate, bits=mixture.bit_depth, channels=mixture.channel_count)
        tfm.set_globals(dither=dither)

        # TODO
        #  Always normalize and remove normalize from list of available augmentations
        #  Normalize to globally set level (should this be a global config parameter,
        #  or hard-coded into the script?)
        if 'normalize' in augmentation:
            tfm.norm(db_level=augmentation['normalize'])

        if 'gain' in augmentation:
            tfm.gain(gain_db=augmentation['gain'], normalize=False, limiter=True)

        if 'pitch' in augmentation:
            tfm.pitch(n_semitones=augmentation['pitch'] / 100)

        if 'tempo' in augmentation:
            factor = augmentation['tempo']
            if abs(factor - 1.0) <= 0.1:
                tfm.stretch(factor=factor)
            else:
                tfm.tempo(factor=factor, audio_type='s')

        if 'eq1' in augmentation:
            tfm.equalizer(frequency=augmentation['eq1'][0], width_q=augmentation['eq1'][1],
                          gain_db=augmentation['eq1'][2])

        if 'eq2' in augmentation:
            tfm.equalizer(frequency=augmentation['eq2'][0], width_q=augmentation['eq2'][1],
                          gain_db=augmentation['eq2'][2])

        if 'eq3' in augmentation:
            tfm.equalizer(frequency=augmentation['eq3'][0], width_q=augmentation['eq3'][1],
                          gain_db=augmentation['eq3'][2])

        if 'lpf' in augmentation:
            tfm.lowpass(frequency=augmentation['lpf'])

        # Create output data
        audio_out = tfm.build_array(input_array=audio_in, sample_rate_in=mixture.sample_rate)

        # make sure length is multiple of length_common_denominator
        audio_len = len(audio_out)
        if audio_len % length_common_denominator:
            pad = length_common_denominator - (audio_len % length_common_denominator)
            audio_out = np.pad(audio_out, pad_width=(0, pad), mode='constant', constant_values=0)

        return audio_out
    except Exception as e:
        logger.exception(f'Error applying {augmentation}: {e}')
        raise SonusAIError



from typing import List

from sonusai import SonusAIError
from sonusai import logger


def get_class_weights_threshold(mixdb: dict) -> List[float]:
    class_weights_threshold = mixdb['class_weights_threshold']
    if not isinstance(class_weights_threshold, list):
        class_weights_threshold = [class_weights_threshold] * mixdb['num_classes']
    if len(class_weights_threshold) != mixdb['num_classes']:
        logger.error(f'invalid class_weights_threshold length: {len(class_weights_threshold)}')
        raise SonusAIError
    return class_weights_threshold



import numpy as np


def calculate_class_weights(truth: np.ndarray, oweight: int = -1, oweight_idx: int = -1) -> np.ndarray:
    # Calculate class weights with support for non-existent classes (a problem with sklearn)
    # where non-existent class gets a weight of 0 (instead of inf)
    # Includes optional weighting of an "other" class if specified
    # ref: cweights = class_weight.compute_class_weight(class_weight='balanced', classes=clabels, y=tlabels)
    # Inputs:
    #   truth    size frames x timesteps x num_classes or frames x num_classes truth data in one-hot format
    #   oweight  weight of the "other" class
    #            >1 will increase weighting/importance relative to the true count
    #            0>oweight<1 will decrease weighting/importance relative
    #            <0 disable, use true count (default = -1)
    #   oweight_idx index of other class in one-hot mode, default = num_classes-1 (the last)

    (frames, num_classes) = truth.shape

    if num_classes > 1:
        # clabels = list(range(0,num_classes))            # num_classes labels 0:num_classes-1
        tlabels = np.argmax(truth, axis=-1)  # framesx1 labels from one-hot, last dim
        cnt = np.bincount(tlabels, minlength=num_classes).astype(float)
    else:
        # clabels = list(range(0,2))        # [0,1] binary case
        num_classes = 2
        tlabels = np.int8(truth >= 0.5)[:, 0]  # quantize to binary and shape (frames,) for bincount
        cnt = np.bincount(tlabels, minlength=num_classes).astype(float)

    if oweight > 0:
        cnt[oweight_idx] = cnt[oweight_idx] / oweight

    cweights = np.empty((len(cnt)))
    for n in range(len(cnt)):
        # avoid sklearn problem with absent classes and assign non-existent class weight of 0
        cweights[n] = frames / (num_classes * cnt[n]) if cnt[n] else 0

    return cweights



import json
from os.path import exists

from sonusai import SonusAIError
from sonusai import logger


def load_mixdb(name: str) -> dict:
    if not exists(name):
        logger.exception(f'{name} does not exist')
        raise SonusAIError

    with open(name, encoding='utf-8') as f:
        mixdb = json.load(f)

    return mixdb



import h5py
import numpy as np

from sonusai import logger


def read_predict_data(filename: str, expected_frames: int) -> (None, np.ndarray):
    if not filename:
        return None

    logger.info(f'Reading prediction data from {filename}')
    with h5py.File(name=filename, mode='r') as f:
        # prediction data is either [frames, num_classes], or [frames, timesteps, num_classes]
        predict = f['predict'][:]

        if predict.ndim == 2:
            frames, num_classes = predict.shape

            if frames != expected_frames:
                logger.warning(f'Ignoring prediction data in {filename} due to frames mismatch')
                return None
        elif predict.ndim == 3:
            frames, timesteps, num_classes = predict.shape

            if frames * timesteps != expected_frames:
                logger.warning(f'Ignoring prediction data in {filename} due to frames mismatch')
                return None

            logger.info(
                f'Reshaping prediction data in {filename} from [{frames}, {timesteps}, {num_classes}] to [{frames * timesteps}, {num_classes}]')
            predict = np.reshape(predict, [frames * timesteps, num_classes], order='F')
        else:
            logger.warning(f'Ignoring prediction data in {filename} due to invalid dimensions')
            return None

        return predict



import yaml

import sonusai
from sonusai import SonusAIError
from sonusai import logger


def get_default_config() -> dict:
    # Load default config
    try:
        with open(sonusai.mixture.default_config, mode='r') as file:
            config = yaml.safe_load(file)
        return config
    except Exception as e:
        logger.exception(f'Error loading genmixdb default config: {e}')
        raise SonusAIError


def get_config_from_file(config_name: str) -> dict:
    config = get_default_config()

    try:
        # Load given config
        with open(config_name, mode='r') as file:
            given_config = yaml.safe_load(file)

        # Use default config as base and overwrite with given config keys as found
        for key in config:
            if key in given_config:
                config[key] = given_config[key]

        return config
    except Exception as e:
        logger.exception(f'Error preparing genmixdb config: {e}')
        raise SonusAIError



import multiprocessing
import random
import tempfile
from multiprocessing import Manager
from typing import List
from typing import Union

import h5py
import numpy as np
from pyaaware import FeatureGenerator
from tensorflow.keras.utils import Sequence

from sonusai import SonusAIError
from sonusai import logger
from sonusai.genft import genft
from sonusai.mixture import get_mixtures_from_mixid
from sonusai.utils import reshape_inputs


class DataGenerator(Sequence):
    """Generates data for Keras"""

    def __init__(self,
                 mixdb: dict,
                 mixid: Union[str, List[int]],
                 batch_size: int,
                 timesteps: int,
                 flatten: bool,
                 add1ch: bool,
                 shuffle: bool = False,
                 chunks: int = 16):
        """Initialization"""
        self.mixdb = mixdb
        self.mixid = mixid
        self.batch_size = batch_size
        self.timesteps = timesteps
        self.flatten = flatten
        self.add1ch = add1ch
        self.shuffle = shuffle
        self.chunks = chunks

        self.index_map = None
        self.total_batches = None
        self.frames_per_batch = None
        self.fetching = multiprocessing.Condition()

        self.cache_file = tempfile.TemporaryFile()
        manager = Manager()
        self.cache_line = manager.list()

        self.initialize_mixtures()

        fg = FeatureGenerator(frame_size=self.mixdb['frame_size'],
                              feature_mode=self.mixdb['feature'],
                              num_classes=self.mixdb['num_classes'],
                              truth_mutex=self.mixdb['truth_mutex'])
        feature, truth, _, _, _, _ = reshape_inputs(
            feature=np.single(np.random.random((self.batch_size, fg.stride, fg.num_bands))),
            truth=np.single(np.random.random((self.batch_size, fg.num_classes,))),
            batch_size=self.batch_size,
            timesteps=self.timesteps,
            flatten=self.flatten,
            add1ch=self.add1ch)

        with h5py.File(self.cache_file, 'a') as f:
            f.create_dataset(name='feature',
                             shape=((self.total_batches, self.batch_size) + feature.shape[1:]),
                             dtype=np.single)
            f.create_dataset(name='truth',
                             shape=((self.total_batches, self.batch_size) + truth.shape[1:]),
                             dtype=np.single)

    def __len__(self):
        """Denotes the number of batches per epoch"""
        return self.total_batches

    def __getitem__(self, batch_index: int):
        """Generate one batch of data"""
        with self.fetching:
            if batch_index not in self.cache_line:
                self.add_data_to_cache(batch_index)

            with h5py.File(self.cache_file, 'r') as f:
                return f['feature'][batch_index], f['truth'][batch_index]

    def on_epoch_end(self):
        """Modification of dataset between epochs"""
        if self.shuffle:
            random.shuffle(self.mixid)
            self.initialize_mixtures()

    def get_feature_frames(self, mixtures: list, mixid: Union[str, List[int]] = ':', start_offset: int = 0) -> int:
        subset = get_mixtures_from_mixid(mixtures, mixid)
        return sum([sub['samples'] for sub in subset]) // self.mixdb['feature_step_samples'] - start_offset

    def initialize_mixtures(self):
        self.mixdb['mixtures'] = get_mixtures_from_mixid(self.mixdb['mixtures'], self.mixid)

        frames = self.get_feature_frames(self.mixdb['mixtures'])
        self.frames_per_batch = self.batch_size if self.timesteps == 0 else self.batch_size * self.timesteps
        self.total_batches = frames // self.frames_per_batch

        if self.total_batches == 0:
            logger.error(
                f'Error: dataset only contains {frames} frames which is not enough to fill a batch size of '
                f'{self.frames_per_batch}. Either provide more data or decrease the batch size')
            raise SonusAIError

        # Compute mixid and offset for dataset
        # offsets are needed because mixtures are not guaranteed to fall on batch boundaries.
        # When fetching a new index that starts in the middle of a sequence of mixtures, the
        # previous feature frame offset must be maintained in order to preserve the correct
        # data sequence.
        self.index_map = list()
        cumulative_frames = 0
        start_mixture_index = 0
        offset = 0
        for mixture_index in range(len(self.mixdb['mixtures'])):
            current_frames = self.mixdb['mixtures'][mixture_index]['samples'] // self.mixdb['feature_step_samples']
            cumulative_frames += current_frames
            while cumulative_frames >= self.frames_per_batch:
                self.index_map.append({
                    'mixid':  list(range(start_mixture_index, mixture_index + 1)),
                    'offset': offset
                })
                extra_frames = cumulative_frames - self.frames_per_batch
                if extra_frames == 0:
                    start_mixture_index = mixture_index + 1
                    offset = 0
                else:
                    start_mixture_index = mixture_index
                    offset = current_frames - extra_frames
                cumulative_frames = extra_frames

    def add_data_to_cache(self, batch_index: int) -> None:
        # Don't skip any indices when filling cache
        if self.cache_line:
            next_cached_index = max(self.cache_line) + 1
            if batch_index > next_cached_index:
                batch_index = next_cached_index

        # Always fetch starting at a chunk boundary
        batch_index = (batch_index // self.chunks) * self.chunks

        mixid = self.index_map[batch_index]['mixid']
        offset = self.index_map[batch_index]['offset']
        chunks = self.get_feature_frames(mixtures=self.mixdb['mixtures'],
                                         mixid=mixid,
                                         start_offset=offset) // self.frames_per_batch
        while chunks < self.chunks and mixid[-1] < len(self.mixdb['mixtures']) - 1:
            mixid.append(mixid[-1] + 1)
            chunks = self.get_feature_frames(mixtures=self.mixdb['mixtures'],
                                             mixid=mixid,
                                             start_offset=offset) // self.frames_per_batch

        feature, truth, _, _ = genft(mixdb=self.mixdb,
                                     mixid=mixid,
                                     logging=False,
                                     parallel=False)

        feature = feature[offset:]
        truth = truth[offset:]

        frames = (feature.shape[0] // self.frames_per_batch) * self.frames_per_batch
        batches = frames // self.frames_per_batch

        if batches == 0:
            logger.error(
                f'Error: genft returned {feature.shape[0]} frames which is not enough to fill a batch size of '
                f'{self.frames_per_batch}. Either provide more data or decrease the batch size')
            raise SonusAIError

        feature = feature[:frames].reshape((batches, self.frames_per_batch, feature.shape[1], feature.shape[2]))
        truth = truth[:frames].reshape((batches, self.frames_per_batch, truth.shape[1]))

        # Add new values to cache file
        with h5py.File(self.cache_file, 'a') as f:
            for index in range(batches):
                self.cache_line.append(batch_index + index)
                new_feature, new_truth, _, _, _, _ = reshape_inputs(feature=feature[index],
                                                                    truth=truth[index],
                                                                    batch_size=self.batch_size,
                                                                    timesteps=self.timesteps,
                                                                    flatten=self.flatten,
                                                                    add1ch=self.add1ch)

                feature_dataset = f['feature']
                feature_dataset[batch_index + index] = new_feature

                truth_dataset = f['truth']
                truth_dataset[batch_index + index] = new_truth



import numpy as np


def averages(metrics: np.ndarray) -> np.ndarray:
    # Compute averages for several averaging methods over classes.
    # For binary these have no meaning macro-avg == statistics of pos. class
    # mavg[0,:]  macro-avg  [PPV, TPR, F1, FPR, ACC, TPSUM]
    # mavg[1,:]  micro-avg  [PPV, TPR, F1, FPR, ACC, TPSUM]  # Note, PPV=TPR=F1=ACC
    # mavg[2,:]  weight-avg [PPV, TPR, F1, FPR, ACC, TPSUM]
    assert metrics.ndim == 2, 'Metrics must be two-dimensional.'
    assert metrics.shape[1] == 12, 'Metrics must have 12 indices.'

    eps = np.finfo(float).eps
    num_classes = np.shape(metrics)[0]
    s = np.sum(metrics[:, 9].astype(int))  # support = sum (true pos total = FN+TP ) over classes

    mavg = np.zeros((3, 6), dtype=np.single)

    # macro average
    mavg[0,] = [np.mean(metrics[:, 2]), np.mean(metrics[:, 1]), np.mean(metrics[:, 6]),
                np.mean(metrics[:, 4]), np.mean(metrics[:, 0]), s]

    # micro average, micro-F1 = micro-precision = micro-recall = accuracy
    if num_classes > 1:
        tp_sum = np.sum(metrics[:, 10])  # TP all classes
        rm = tp_sum / (np.sum(metrics[:, 9]) + eps)  # micro mean PPV = TP / (PT=FN+TP)
        fp_sum = np.sum(metrics[:, 11])  # FP all classes
        fpm = fp_sum / (np.sum(metrics[:, 8]) + eps)  # micro mean FPR = FP / (NT=TN+FP)
        # pm  = tp_sum / (tp_sum + fp_sum + eps)      # micro mean TPR = TP / (TP+FP) (note: same as rm for micro-avg)
        mavg[1,] = [rm, rm, rm, fpm, rm, s]  # specific format, last 3 are unique
        # weighted average TBD
        # mavg[2,:] =
    else:  # binary case, all are same
        mavg[1,] = mavg[0,]
        mavg[2,] = mavg[0,]

    return mavg



"""genft

usage: genft [-hvs] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture IDs JSON file.
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI feature/truth file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A JSON file containing a list of mixture IDs. The list should be named 'mixid'.
                If no file is provided, then all mixtures in the database will be generated.

Outputs:
    OUTPUT.h5   A SonusAI feature HDF5 file. Contains:
                    dataset:    feature
                    dataset:    truth_f
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genft.log

"""

import ctypes
import json
import multiprocessing
import time
from os.path import splitext
from typing import List
from typing import Union

import h5py
import numpy as np
from docopt import docopt
from pyaaware import FeatureGenerator
from pyaaware import ForwardTransform
from tqdm import tqdm

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import audio_files_exist
from sonusai.mixture import build_noise_audio_db
from sonusai.mixture import build_target_audio_db
from sonusai.mixture import get_audio_and_truth_t
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_mixdb
from sonusai.mixture import load_mixid
from sonusai.mixture import new_mixdb_from_mixid
from sonusai.mixture import set_mixture_offsets
from sonusai.mixture import truth_reduction
from sonusai.utils import grouper
from sonusai.utils import human_readable_size
from sonusai.utils import int16_to_float
from sonusai.utils import p_map
from sonusai.utils import seconds_to_hms
from sonusai.utils import to_numpy_array
from sonusai.utils import trim_docstring

genft_global = dict()


# noinspection PyGlobalUndefined

def mp_init(feature_: multiprocessing.Array,
            truth_f_: multiprocessing.Array,
            segsnr_: multiprocessing.Array) -> None:
    global mp_feature
    global mp_truth_f
    global mp_segsnr

    mp_feature = feature_
    mp_truth_f = truth_f_
    mp_segsnr = segsnr_


def init(mixdb: dict,
         mixid: Union[str, List[int]],
         logging: bool = True) -> (dict, int, int, int):
    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub['samples'] for sub in mixdb_out['mixtures']])
    transform_frames = total_samples // mixdb_out['frame_size']
    feature_frames = total_samples // mixdb_out['feature_step_samples']

    if logging:
        logger.info('')
        logger.info(f'Found {len(mixdb_out["mixtures"])} mixtures to process')
        logger.info(f'{total_samples} samples, {transform_frames} transform frames, {feature_frames} feature frames')

    return mixdb_out, total_samples, transform_frames, feature_frames


def genft(mixdb: dict,
          mixid: Union[str, List[int]],
          compute_segsnr: bool = False,
          logging: bool = True,
          show_progress: bool = False,
          progress: tqdm = None,
          parallel: bool = True) -> (np.ndarray, np.ndarray, dict):
    mixdb_out, total_samples, transform_frames, feature_frames = init(mixdb=mixdb,
                                                                      mixid=mixid,
                                                                      logging=logging)

    genft_global['mixdb'] = mixdb_out
    genft_global['feature_frames'] = feature_frames
    genft_global['compute_segsnr'] = compute_segsnr
    genft_global['noise_audios'] = build_noise_audio_db(mixdb=mixdb_out, show_progress=show_progress)
    genft_global['target_audios'] = build_target_audio_db(mixdb=mixdb_out, show_progress=show_progress)

    fg = FeatureGenerator(frame_size=mixdb_out['frame_size'],
                          feature_mode=mixdb_out['feature'],
                          num_classes=mixdb_out['num_classes'],
                          truth_mutex=mixdb_out['truth_mutex'])
    mp_feature = multiprocessing.Array(ctypes.c_float, feature_frames * fg.stride * fg.num_bands)
    mp_truth_f = multiprocessing.Array(ctypes.c_float, feature_frames * fg.num_classes)
    mp_segsnr = multiprocessing.Array(ctypes.c_float, 0)
    if compute_segsnr:
        mp_segsnr = multiprocessing.Array(ctypes.c_float, transform_frames)

    feature = to_numpy_array(mp_feature, dtype=np.single).reshape((feature_frames, fg.stride, fg.num_bands))
    truth_f = to_numpy_array(mp_truth_f, dtype=np.single).reshape((feature_frames, fg.num_classes))
    segsnr = np.empty(0, dtype=np.single)
    if compute_segsnr:
        segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    # First pass to get offsets
    set_mixture_offsets(mixdb_out)

    # Second pass to get feature and truth_f
    progress_needs_close = False
    if progress is None:
        progress = tqdm(total=len(mixdb_out['mixtures']), desc='genft', disable=not show_progress)
        progress_needs_close = True

    if parallel:
        p_map(process_mixture,
              mixdb_out['mixtures'],
              progress=progress,
              initializer=mp_init,
              initargs=(mp_feature, mp_truth_f, mp_segsnr))
    else:
        mp_init(mp_feature, mp_truth_f, mp_segsnr)
        for record in mixdb_out['mixtures']:
            process_mixture(record)
            progress.update()

    if progress_needs_close:
        progress.close()

    mixdb_out['class_count'] = get_total_class_count(mixdb_out)

    duration = total_samples / sonusai.mixture.sample_rate
    if logging:
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')
        logger.info(f'truth_f:  {human_readable_size(truth_f.nbytes, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')

    return feature, truth_f, segsnr, mixdb_out


def process_mixture(record: dict) -> None:
    fft = ForwardTransform(N=genft_global['mixdb']['frame_size'] * 4,
                           R=genft_global['mixdb']['frame_size'])

    fg = FeatureGenerator(frame_size=genft_global['mixdb']['frame_size'],
                          feature_mode=genft_global['mixdb']['feature'],
                          num_classes=genft_global['mixdb']['num_classes'],
                          truth_mutex=genft_global['mixdb']['truth_mutex'])

    feature = to_numpy_array(mp_feature, dtype=np.single)
    feature = feature.reshape((genft_global['feature_frames'], fg.stride, fg.num_bands))

    truth_f = to_numpy_array(mp_truth_f, dtype=np.single)
    truth_f = truth_f.reshape((genft_global['feature_frames'], fg.num_classes))

    segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    frame_indices = slice(record['i_frame_offset'],
                          record['i_frame_offset'] + record['samples'] // genft_global['mixdb']['frame_size'])

    (mixture_td,
     truth_t,
     _,
     _,
     segsnr[frame_indices]) = get_audio_and_truth_t(mixdb=genft_global['mixdb'],
                                                    record=record,
                                                    target_audios=genft_global['target_audios'],
                                                    noise_audios=genft_global['noise_audios'],
                                                    compute_truth=True,
                                                    compute_segsnr=genft_global['compute_segsnr'],
                                                    frame_based_segnsr=True)

    transform_frames = record['samples'] // genft_global['mixdb']['frame_size']
    feature_frame = record['o_frame_offset']
    for transform_frame in range(transform_frames):
        indices = slice(transform_frame * genft_global['mixdb']['frame_size'],
                        (transform_frame + 1) * genft_global['mixdb']['frame_size'])
        mixture_fd = fft.execute(int16_to_float(mixture_td[indices]))

        fg.execute(mixture_fd, truth_reduction(truth_t[indices], genft_global['mixdb']['truth_reduction_function']))

        if fg.eof():
            feature[feature_frame] = fg.feature()
            truth_f[feature_frame] = fg.truth()
            feature_frame += 1


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        mixdb_name = args['--mixdb']
        mixid_name = args['--mixid']
        output_name = args['--output']
        compute_segsnr = args['--segsnr']

        if not output_name:
            output_name = splitext(mixdb_name)[0] + '.h5'

        start_time = time.monotonic()

        log_name = 'genft.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('genft')

        mixdb = load_mixdb(name=mixdb_name)
        mixid = load_mixid(name=mixid_name, mixdb=mixdb)

        mixdb_out, total_samples, transform_frames, feature_frames = init(mixdb=mixdb,
                                                                          mixid=mixid,
                                                                          logging=True)

        if not audio_files_exist(mixdb):
            raise SystemExit(1)

        fg = FeatureGenerator(frame_size=mixdb_out['frame_size'],
                              feature_mode=mixdb_out['feature'],
                              num_classes=mixdb_out['num_classes'],
                              truth_mutex=mixdb_out['truth_mutex'])

        with h5py.File(output_name, 'w') as f:
            f.create_dataset(name='feature',
                             shape=(feature_frames, fg.stride, fg.num_bands),
                             dtype=np.single)
            f.create_dataset(name='truth_f',
                             shape=(feature_frames, fg.num_classes),
                             dtype=np.single)
            if compute_segsnr:
                f.create_dataset(name='segsnr',
                                 shape=(transform_frames,),
                                 dtype=np.single)

        chunk_size = 100
        progress = tqdm(total=len(mixid), desc='genft')
        mixid = grouper(range(len(mixdb_out['mixtures'])), chunk_size)
        mixdb_out['class_count'] = [0] * mixdb_out['num_classes']

        i_frame_offset = 0
        o_frame_offset = 0
        for m in mixid:
            feature, truth_f, segsnr, mixdb_tmp = genft(mixdb=mixdb_out,
                                                        mixid=m,
                                                        compute_segsnr=compute_segsnr,
                                                        logging=False,
                                                        progress=progress)
            o_frames = feature.shape[0]
            if o_frames != truth_f.shape[0]:
                logger.exception(f'truth_f frames does not match feature frames: {truth_f.shape[0]} != {o_frames}')
                raise SystemExit(1)

            with h5py.File(output_name, 'a') as f:
                indices = slice(o_frame_offset, o_frame_offset + o_frames)
                o_frame_offset += o_frames

                feature_dataset = f['feature']
                feature_dataset[indices] = feature

                truth_dataset = f['truth_f']
                truth_dataset[indices] = truth_f

                if compute_segsnr:
                    segsnr_dataset = f['segsnr']
                    segsnr_dataset[i_frame_offset:i_frame_offset + segsnr.shape[0]] = segsnr
                    i_frame_offset += segsnr.shape[0]

            for idx, val in enumerate(m):
                mixdb_out['mixtures'][val] = mixdb_tmp['mixtures'][idx]
            for idx in range(mixdb_out['num_classes']):
                mixdb_out['class_count'][idx] += mixdb_tmp['class_count'][idx]

            with h5py.File(output_name, 'a') as f:
                f.attrs['mixdb'] = json.dumps(mixdb_out)

        progress.close()

        logger.info(f'Wrote {output_name}')
        duration = total_samples / sonusai.mixture.sample_rate
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'feature:  {human_readable_size(o_frame_offset * fg.stride * fg.num_bands * 4, 1)}')
        logger.info(f'truth_f:  {human_readable_size(o_frame_offset * fg.num_classes * 4, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(i_frame_offset * 4, 1)}')

        end_time = time.monotonic()
        logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import os
import string


def expandvars(path: str) -> (str, dict):
    """Expand shell variables of the forms $var, ${var} and %var%.
    Unknown variables are left unchanged."""

    # Expand paths containing shell variable substitutions.
    # The following rules apply:
    #       - no expansion within single quotes
    #       - '$$' is translated into '$'
    #       - '%%' is translated into '%' if '%%' are not seen in %var1%%var2%
    #       - ${varname} is accepted.
    #       - $varname is accepted.
    #       - %varname% is accepted.
    #       - varnames can be made out of letters, digits and the characters '_-'
    #         (though is not verified in the ${varname} and %varname% cases)
    # XXX With COMMAND.COM you can use any characters in a variable name,
    # XXX except '^|<>='.
    path = os.fspath(path)
    token_map = {}
    if isinstance(path, bytes):
        if b'$' not in path and b'%' not in path:
            return path, token_map
        varchars = bytes(string.ascii_letters + string.digits + '_-', 'ascii')
        quote = b'\''
        percent = b'%'
        brace = b'{'
        rbrace = b'}'
        dollar = b'$'
        environ = getattr(os, 'environb', None)
    else:
        if '$' not in path and '%' not in path:
            return path, token_map
        varchars = string.ascii_letters + string.digits + '_-'
        quote = '\''
        percent = '%'
        brace = '{'
        rbrace = '}'
        dollar = '$'
        environ = os.environ
    res = path[:0]
    index = 0
    pathlen = len(path)
    while index < pathlen:
        c = path[index:index + 1]
        if c == quote:  # no expansion within single quotes
            path = path[index + 1:]
            pathlen = len(path)
            try:
                index = path.index(c)
                res += c + path[:index + 1]
            except ValueError:
                res += c + path
                index = pathlen - 1
        elif c == percent:  # variable or '%'
            if path[index + 1:index + 2] == percent:
                res += c
                index += 1
            else:
                path = path[index + 1:]
                pathlen = len(path)
                try:
                    index = path.index(percent)
                except ValueError:
                    res += percent + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)])
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = percent + var + percent
                    res += value
        elif c == dollar:  # variable or '$$'
            if path[index + 1:index + 2] == dollar:
                res += c
                index += 1
            elif path[index + 1:index + 2] == brace:
                path = path[index + 2:]
                pathlen = len(path)
                try:
                    index = path.index(rbrace)
                except ValueError:
                    res += dollar + brace + path
                    index = pathlen - 1
                else:
                    var = path[:index]
                    try:
                        if environ is None:
                            value = os.fsencode(os.environ[os.fsdecode(var)])
                        else:
                            value = environ[var]
                        token_map[var] = value
                    except KeyError:
                        value = dollar + brace + var + rbrace
                    res += value
            else:
                var = path[:0]
                index += 1
                c = path[index:index + 1]
                while c and c in varchars:
                    var += c
                    index += 1
                    c = path[index:index + 1]
                try:
                    if environ is None:
                        value = os.fsencode(os.environ[os.fsdecode(var)])
                    else:
                        value = environ[var]
                    token_map[var] = value
                except KeyError:
                    value = dollar + var
                res += value
                if c:
                    index -= 1
        else:
            res += c
        index += 1
    return res, token_map



from sonusai import SonusAIError
from sonusai import logger


def get_offsets(mixdb: dict, mixid: int) -> (int, int, int):
    required_keys = [
        'mixtures',
        'frame_size',
        'feature_step_samples',
    ]
    for key in required_keys:
        if key not in mixdb.keys():
            logger.exception(f'Missing {key} in mixdb')
            raise SonusAIError

    if mixid >= len(mixdb['mixtures']) or mixid < 0:
        logger.exception(f'Invalid mixid: {mixid}')
        raise SonusAIError

    i_sample_offset = sum([sub['samples'] for sub in mixdb['mixtures'][:mixid + 1]])
    i_frame_offset = i_sample_offset // mixdb['frame_size']
    o_frame_offset = i_sample_offset // mixdb['feature_step_samples']

    return i_sample_offset, i_frame_offset, o_frame_offset



import numpy as np


def confusion_matrix_to_string(cm: np.ndarray, normalize: bool = False) -> str:
    # Convert confusion matrix to string
    # If normalize is False (default) then print standard non-normalized confusion matrix,
    # else print normalized confusion matrix
    assert cm.ndim == 2, 'Confusion matrix must be two-dimensional.'
    assert cm.shape[0] == cm.shape[1], 'Confusion matrix must be square.'

    num_classes = np.shape(cm)[0]
    row_count = np.sum(cm, 1)
    total_count = np.sum(row_count)

    result = ''
    if not normalize:
        result += f'Confusion matrix with num_classes, total count: {num_classes},{total_count}:\n'
        with np.printoptions(suppress=True):
            result += f'{cm.astype(int)}\n'
    else:
        eps = np.finfo(float).eps
        norm_val = (np.tile(np.reshape(row_count, (num_classes, 1)), (1, num_classes)) + eps)
        ncm = cm / norm_val  # element by element
        result += f'Normalized confusion matrix (%) with num_classes, total count: {num_classes},{total_count}:\n'
        with np.printoptions(precision=2, suppress=True):
            result += f'{np.round(ncm * 100, 0).astype(int)}\n'

    return result



"""genmix

usage: genmix [-hvts] (-d MIXDB) [-i MIXID] [-o OUTPUT]

options:
    -h, --help
    -v, --verbose                   Be verbose.
    -d MIXDB, --mixdb MIXDB         Mixture database JSON file.
    -i MIXID, --mixid MIXID         Mixture IDs JSON file.
    -o OUTPUT, --output OUTPUT      Output HDF5 file.
    -t, --truth                     Save truth_t. [default: False].
    -s, --segsnr                    Save segsnr. [default: False].

Generate a SonusAI mixture file from a SonusAI mixture database.

Inputs:
    MIXDB       A SonusAI mixture database JSON file.
    MIXID       A JSON file containing a list of mixture IDs. The list should be named 'mixid'.
                If no file is provided, then all mixtures in the database will be generated.

Outputs:
    OUTPUT.h5   A SonusAI mixture HDF5 file. Contains:
                    dataset:    mixture
                    dataset:    truth_t (optional)
                    dataset:    target
                    dataset:    noise
                    dataset:    segsnr (optional)
                    attribute:  mixdb
    genmix.log
"""
import ctypes
import json
import multiprocessing
import time
from os.path import splitext
from typing import List
from typing import Union

import h5py
import numpy as np
from docopt import docopt
from tqdm import tqdm

import sonusai
from sonusai import SonusAIError
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.mixture import apply_augmentation
from sonusai.mixture import audio_files_exist
from sonusai.mixture import build_noise_audio_db
from sonusai.mixture import build_target_audio_db
from sonusai.mixture import get_audio_and_truth_t
from sonusai.mixture import get_total_class_count
from sonusai.mixture import load_mixdb
from sonusai.mixture import load_mixid
from sonusai.mixture import new_mixdb_from_mixid
from sonusai.mixture import set_mixture_offsets
from sonusai.utils import grouper
from sonusai.utils import human_readable_size
from sonusai.utils import p_map
from sonusai.utils import seconds_to_hms
from sonusai.utils import to_numpy_array
from sonusai.utils import trim_docstring

genmix_global = dict()


# noinspection PyGlobalUndefined

def mp_init(mixture_: multiprocessing.Array,
            truth_t_: multiprocessing.Array,
            target_: multiprocessing.Array,
            noise_: multiprocessing.Array,
            segsnr_: multiprocessing.Array) -> None:
    global mp_mixture
    global mp_truth_t
    global mp_target
    global mp_noise
    global mp_segsnr

    mp_mixture = mixture_
    mp_truth_t = truth_t_
    mp_target = target_
    mp_noise = noise_
    mp_segsnr = segsnr_


def init(mixdb: dict,
         mixid: Union[str, List[int]],
         logging: bool = True) -> (dict, int):
    mixdb_out = new_mixdb_from_mixid(mixdb=mixdb, mixid=mixid)

    total_samples = sum([sub['samples'] for sub in mixdb_out['mixtures']])

    if logging:
        logger.info('')
        logger.info(f'Found {len(mixdb_out["mixtures"])} mixtures to process')
        logger.info(f'{total_samples} samples')

    return mixdb_out, total_samples


def genmix(mixdb: dict,
           mixid: Union[str, List[int]],
           compute_segsnr: bool = False,
           compute_truth: bool = False,
           logging: bool = False,
           show_progress: bool = False,
           progress: tqdm = None) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, dict):
    mixdb_out, total_samples = init(mixdb=mixdb, mixid=mixid, logging=logging)

    genmix_global['mixdb'] = mixdb_out
    genmix_global['total_samples'] = total_samples
    genmix_global['compute_truth'] = compute_truth
    genmix_global['compute_segsnr'] = compute_segsnr
    genmix_global['noise_audios'] = build_noise_audio_db(mixdb_out)
    genmix_global['target_audios'] = build_target_audio_db(mixdb_out)

    mp_mixture = multiprocessing.Array(ctypes.c_int16, total_samples)
    mp_truth_t = multiprocessing.Array(ctypes.c_float, 0)
    if compute_truth:
        mp_truth_t = multiprocessing.Array(ctypes.c_float, total_samples * mixdb_out['num_classes'])
    mp_target = multiprocessing.Array(ctypes.c_int16, total_samples)
    mp_noise = multiprocessing.Array(ctypes.c_int16, total_samples)
    mp_segsnr = multiprocessing.Array(ctypes.c_float, 0)
    if compute_segsnr:
        mp_segsnr = multiprocessing.Array(ctypes.c_float, total_samples)

    mixture = to_numpy_array(mp_mixture, dtype=np.int16)
    truth_t = np.empty(0, dtype=np.single)
    if compute_truth:
        truth_t = to_numpy_array(mp_truth_t, dtype=np.single).reshape((total_samples, mixdb_out['num_classes']))
    target = to_numpy_array(mp_target, dtype=np.int16)
    noise = to_numpy_array(mp_noise, dtype=np.int16)
    segsnr = np.empty(0, dtype=np.single)
    if compute_segsnr:
        segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    # First pass to get offsets
    set_mixture_offsets(mixdb_out)

    # Second pass to get mixture and truth_t
    progress_needs_close = False
    if progress is None:
        progress = tqdm(total=len(mixdb_out['mixtures']), desc='genmix', disable=not show_progress)
        progress_needs_close = True

    p_map(process_mixture, mixdb_out['mixtures'], progress=progress, initializer=mp_init, initargs=(mp_mixture,
                                                                                                    mp_truth_t,
                                                                                                    mp_target,
                                                                                                    mp_noise,
                                                                                                    mp_segsnr))

    if progress_needs_close:
        progress.close()

    mixdb_out['class_count'] = get_total_class_count(mixdb_out)

    duration = len(mixture) / sonusai.mixture.sample_rate
    if logging:
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'mixture:  {human_readable_size(mixture.nbytes, 1)}')
        if compute_truth:
            logger.info(f'truth_t:  {human_readable_size(truth_t.nbytes, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')

    return mixture, truth_t, target, noise, segsnr, mixdb_out


def process_mixture(record: dict) -> None:
    indices = slice(record['i_sample_offset'], record['i_sample_offset'] + record['samples'])
    mixture = to_numpy_array(mp_mixture, dtype=np.int16)
    truth_t = to_numpy_array(mp_truth_t, dtype=np.single)
    if genmix_global['compute_truth']:
        truth_t = truth_t.reshape((genmix_global['total_samples'], genmix_global['mixdb']['num_classes']))
    target = to_numpy_array(mp_target, dtype=np.int16)
    noise = to_numpy_array(mp_noise, dtype=np.int16)
    segsnr = to_numpy_array(mp_segsnr, dtype=np.single)

    (mixture[indices],
     truth_t[indices],
     target[indices],
     noise[indices],
     segsnr[indices]) = get_audio_and_truth_t(mixdb=genmix_global['mixdb'],
                                              record=record,
                                              target_audios=genmix_global['target_audios'],
                                              noise_audios=genmix_global['noise_audios'],
                                              compute_truth=genmix_global['compute_truth'],
                                              compute_segsnr=genmix_global['compute_segsnr'])


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        verbose = args['--verbose']
        mixdb_name = args['--mixdb']
        mixid_name = args['--mixid']
        output_name = args['--output']
        compute_segsnr = args['--segsnr']
        compute_truth = args['--truth']

        if not output_name:
            output_name = splitext(mixdb_name)[0] + '.h5'

        start_time = time.monotonic()

        log_name = 'genmix.log'
        create_file_handler(log_name)
        update_console_handler(verbose)
        initial_log_messages('genmix')

        mixdb = load_mixdb(name=mixdb_name)
        mixid = load_mixid(name=mixid_name, mixdb=mixdb)

        mixdb_out, total_samples = init(mixdb=mixdb,
                                        mixid=mixid,
                                        logging=True)

        if not audio_files_exist(mixdb):
            raise SonusAIError

        with h5py.File(output_name, 'w') as f:
            mixture_dataset = f.create_dataset(name='mixture',
                                               shape=(total_samples,),
                                               dtype=np.int16)
            if compute_truth:
                truth_dataset = f.create_dataset(name='truth_t',
                                                 shape=(total_samples, mixdb_out['num_classes']),
                                                 dtype=np.single)
            target_dataset = f.create_dataset(name='target',
                                              shape=(total_samples,),
                                              dtype=np.int16)
            noise_dataset = f.create_dataset(name='noise',
                                             shape=(total_samples,),
                                             dtype=np.int16)
            if compute_segsnr:
                segsnr_dataset = f.create_dataset(name='segsnr',
                                                  shape=(total_samples,),
                                                  dtype=np.single)
        chunk_size = 100
        progress = tqdm(total=len(mixid), desc='genmix')
        mixid = grouper(range(len(mixdb_out['mixtures'])), chunk_size)
        mixdb_out['class_count'] = [0] * mixdb_out['num_classes']

        sample_offset = 0
        chunk_offset = 0
        for m in mixid:
            mixture, truth_t, target, noise, segsnr, mixdb_tmp = genmix(mixdb=mixdb_out,
                                                                        mixid=m,
                                                                        compute_segsnr=compute_segsnr,
                                                                        compute_truth=compute_truth,
                                                                        logging=False,
                                                                        progress=progress)

            samples = mixture.shape[0]
            if compute_truth:
                if samples != truth_t.shape[0]:
                    logger.exception(
                        f'truth_t samples does not match mixture samples: {truth_t.shape[0]} != {samples}')
                    raise SonusAIError
                if mixdb_out['num_classes'] != truth_t.shape[1]:
                    logger.exception(
                        f'truth_t num_classes is incorrect: {truth_t.shape[1]} != {mixdb_out["num_classes"]}')
                    raise SonusAIError
            if samples != target.shape[0]:
                logger.exception(f'target samples does not match mixture samples: {target.shape[0]} != {samples}')
                raise SonusAIError
            if samples != noise.shape[0]:
                logger.exception(f'noise samples does not match mixture samples: {noise.shape[0]} != {samples}')
                raise SonusAIError
            if compute_segsnr and samples != segsnr.shape[0]:
                logger.exception(f'segsnr samples does not match mixture samples: {segsnr.shape[0]} != {samples}')
                raise SonusAIError

            with h5py.File(output_name, 'a') as f:
                indices = slice(sample_offset, sample_offset + samples)
                sample_offset += samples

                mixture_dataset = f['mixture']
                mixture_dataset[indices] = mixture

                if compute_truth:
                    truth_dataset = f['truth_t']
                    truth_dataset[indices] = truth_t

                target_dataset = f['target']
                target_dataset[indices] = target

                noise_dataset = f['noise']
                noise_dataset[indices] = noise

                if compute_segsnr:
                    segsnr_dataset = f['segsnr']
                    segsnr_dataset[indices] = segsnr

            for idx, val in enumerate(m):
                mixdb_out['mixtures'][val] = mixdb_tmp['mixtures'][idx]
            for idx in range(mixdb_out['num_classes']):
                mixdb_out['class_count'][idx] += mixdb_tmp['class_count'][idx]
            chunk_offset += chunk_size

        with h5py.File(output_name, 'a') as f:
            f.attrs['mixdb'] = json.dumps(mixdb_out)

        progress.close()

        logger.info(f'Wrote {output_name}')
        duration = total_samples / sonusai.mixture.sample_rate
        logger.info('')
        logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
        logger.info(f'mixture:  {human_readable_size(sample_offset * 2, 1)}')
        if compute_truth:
            logger.info(f'truth_t:  {human_readable_size(sample_offset * mixdb_out["num_classes"] * 4, 1)}')
        logger.info(f'target:   {human_readable_size(sample_offset * 2, 1)}')
        logger.info(f'noise:    {human_readable_size(sample_offset * 2, 1)}')
        if compute_segsnr:
            logger.info(f'segsnr:   {human_readable_size(sample_offset * 4, 1)}')

        end_time = time.monotonic()
        logger.info(f'Completed in {seconds_to_hms(seconds=end_time - start_time)}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



from typing import List

import numpy as np


def format_confusion_matrix(cm: np.ndarray, labels: List[str] = None) -> str:
    # print confusion matrix from float variable cm and optional labels list
    return f'{cm.astype(int)}\n'



"""evaluate

usage: evaluate [-hv] (-f FEATURE) [-m MIXTURE] [-n PREDICT] [-b BTHR] [-p PLOTNUM]

options:
   -h, --help
   -v, --verbose                    Be verbose.
   -f FEATURE, --feature FEATURE    Feature .h5 data file.
   -m MIXTURE, --mixture MIXTURE    Optional mixture .h5 data file.
   -n PREDICT, --predict PREDICT    Optional predict .h5 data file.
   -b BTHR, --bthr BTHR             Optional binary detection threshold, 0 = argmax(). [default: 0].
   -p PLOTNUM, --plotnum PLOTNUM    Optional plot mixture results (-1 is plot all, 0 is plot none) [default: 0].

The evaluate command measures performance of neural-network models from frame-by-frame truth and prediction data.
It supports data mixtures of target and noise files at various SNR levels as created by sonusai genft function.

Inputs:
    FEATURE     A SonusAI feature HDF5 file. Contains:
                    dataset:    feature
                    dataset:    truth_f
                    dataset:    segsnr
                    attribute:  mixdb
    MIXTURE     A SonusAI mixture HDF5 file. Contains:
                    dataset:    mixture
                    dataset:    target
                    dataset:    noise
    PREDICT     A SonusAI predict HDF5 file. Contains:
                    dataset:    predict (either [frames, num_classes] or [frames, timesteps, num_classes])
"""
from datetime import datetime
from os import mkdir
from os.path import basename
from os.path import splitext
from pathlib import Path
from typing import Union

import numpy as np
from docopt import docopt
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.metrics import averages
from sonusai.metrics import calculate_metrics
from sonusai.metrics import format_confusion_matrix
from sonusai.metrics import generate_snr_summary
from sonusai.metrics import generate_summary
from sonusai.metrics import one_hot
from sonusai.utils import int16_to_float
from sonusai.utils import read_feature_data
from sonusai.utils import read_mixture_data
from sonusai.utils import read_predict_data
from sonusai.utils import trim_docstring


def evaluate(mixdb: dict,
             feature: np.ndarray,
             truth: np.ndarray,
             segsnr: np.ndarray,
             output_dir: str,
             mixture: Union[None, np.ndarray] = None,
             target: Union[None, np.ndarray] = None,
             noise: Union[None, np.ndarray] = None,
             predict: Union[None, np.ndarray] = None,
             binary_detection_threshold: float = 0.0,
             plot_number: int = 0,
             verbose: bool = False):
    update_console_handler(verbose)
    initial_log_messages('evaluate')

    # Summarize the mixture data
    # Number of mixtures
    num_mixtures = len(mixdb['mixtures'])

    if plot_number < 0:
        # If plot_number is negative, plot all
        plot_number = [*range(num_mixtures)]
    elif plot_number == 0:
        # If plot_number is zero, plot none
        plot_number = []
    else:
        # plot_number needs to be a list (for looping later)
        plot_number = [plot_number]

    # Number of feature frames
    feature_frames = feature.shape[0]

    # Number of samples per feature
    samples_per_feature = mixdb['feature_samples']

    # Total mixture audio samples
    mixture_samples = feature_frames * samples_per_feature

    # Number of transform frames
    transform_frames = segsnr.shape[0]

    # Number of samples per transform frame
    samples_per_transform = int(mixture_samples / transform_frames)

    # Number of classes
    num_classes = truth.shape[1]

    # Summarize mixture data by noise file, target file, SNR via array mstat:
    # size NNFxNTFxNSNRxNAUGx20
    #  #noise-files, #SNR, #target-files, #augmentations, #param fields.
    # We track all other augmentations in a dimension that auto-increments, so we can support any combinations.
    # param fields:
    #          mi,target_snr,...
    noise_files = mixdb['noise_files']
    num_noise_files = len(noise_files)
    target_files = mixdb['target_files']
    num_target_files = len(target_files)

    # Determine number of SNR cases
    snrs = []
    for target_augmentation in mixdb['target_augmentations']:
        if 'snr' in target_augmentation.keys():
            snrs.append(target_augmentation['snr'])
    snrs = sorted(list(set(snrs)), reverse=True)

    num_snrs = len(snrs)  # Number of SNR cases
    # Detect special noise-only cases, specified by SNR < -96 where genft sets target gain = 0
    np_snrs = []
    for i, x in enumerate(snrs):
        if float(x) < -96:
            np_snrs.append(i)
    # Index of noise performance SNR (should only be 1 of them)
    if len(np_snrs) > 1:
        logger.debug('Error: more than one noise performance SNR < -96 dB detected, proceeding with first one.')
        np_snrs = [np_snrs[0]]

    num_augmentation_types = num_mixtures / (num_noise_files * num_target_files * num_snrs)
    if num_augmentation_types % 1:  # Number of augmentation types should always be integer
        logger.warning('Number of augmentations is fractional.')
    num_augmentation_types = int(np.ceil(num_augmentation_types))

    logger.info(f'Detected {num_augmentation_types} augmentation types beyond noise, target, and SNR.')

    if mixdb['truth_mutex']:
        logger.debug('Detected truth mutex mode')
        truth_is_mutex = True
    else:
        truth_is_mutex = False

    logger.info(
        f'Analyzing performance for {num_mixtures} mixtures, {num_noise_files} noise files, {num_target_files} target files, and {num_snrs} SNR levels.')

    if binary_detection_threshold == 0:
        binary_detection_threshold = 0.5
        cmode = 0
        logger.debug(
            f'Using default argmax() for detection (binary binary_detection_threshold = {binary_detection_threshold:4.2f})')
    else:
        cmode = binary_detection_threshold
        logger.debug(f'Using custom detection threshold of {binary_detection_threshold:4.2f}')

    # Mixture stats
    mstat = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 23), dtype=np.single)
    # Multiclass stats
    mcmall = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, num_classes, 2, 2),
                      dtype=np.single)
    if num_classes > 1:
        cmall = np.zeros(
            shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, num_classes, num_classes),
            dtype=np.single)
    else:
        cmall = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 2, 2),
                         dtype=np.single)

    # Keep indices in fields [gsnr,HITFA,FPR,TPR,segsnravg,mxst.segsnrpkmed,mxst.NF]
    fnameidx = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 5), dtype=np.int16)

    # Initial augmentation type
    augi = 0
    sub_predict = []

    miwidth = len(str(num_mixtures - 1))
    tfwidth = max(len(ele) for ele in [Path(x).stem for x in target_files])
    nfwidth = max(len(ele) for ele in [Path(x).stem for x in noise_files])

    if plot_number:
        pdf = PdfPages(f'{output_dir}/evaluate.pdf')

    for mi in range(num_mixtures):
        ni = mixdb['mixtures'][mi]['noise_file_index']
        ti = mixdb['mixtures'][mi]['target_file_index']

        # Get mixture config parameters target gain (and target SNR above)
        target_augmentation_index = mixdb['mixtures'][mi]['target_augmentation_index']
        target_snr = mixdb['target_augmentations'][target_augmentation_index]['snr']

        if 'gain' in mixdb['target_augmentations'][target_augmentation_index].keys():
            target_gain = mixdb['target_augmentations'][target_augmentation_index]['gain']
        else:
            target_gain = 1

        target_snr_index = snrs.index(target_snr)

        # record the truth index, provided by genft per mixture.
        truth_index = mixdb['mixtures'][mi]['truth_index'] - 1
        # store index into filename lists in case we want full filenames
        fnameidx[ni, ti, target_snr_index, augi,] = [ni, ti, truth_index, target_snr_index, augi]

        # For each mixture, get index endpoints (mix,noise,target sample,...)
        i_sample_begin = mixdb['mixtures'][mi]['i_sample_offset']
        i_frame_begin = mixdb['mixtures'][mi]['i_frame_offset']
        o_frame_begin = mixdb['mixtures'][mi]['o_frame_offset']
        if mi == num_mixtures - 1:
            i_sample_end = mixture_samples
            i_frame_end = transform_frames
            o_frame_end = feature_frames
        else:
            i_sample_end = mixdb['mixtures'][mi + 1]['i_sample_offset']
            # in frames are transform
            i_frame_end = mixdb['mixtures'][mi + 1]['i_frame_offset']
            # out frames are possibly decimated/stride-combined
            o_frame_end = mixdb['mixtures'][mi + 1]['o_frame_offset']

        # Select subset of waveforms, if present
        sub_mixture = []
        if mixture is not None:
            sub_mixture = int16_to_float(mixture[i_sample_begin:i_sample_end])
        sub_noise = []
        if noise is not None:
            sub_noise = int16_to_float(noise[i_sample_begin:i_sample_end])
        sub_target = []
        if target is not None:
            sub_target = int16_to_float(target[i_sample_begin:i_sample_end])

        noise_level = get_level(sub_noise)
        target_level = get_level(sub_target)

        mixture_snr = target_level - noise_level

        # Segmental SNR and target stats
        sub_segnsr = segsnr[i_frame_begin:i_frame_end]

        # Replace float('Inf') with weighted filter of 16 past samples, then median filter
        sub_segnsr = segsnr_filter(sub_segnsr)

        # segmental SNR mean = mixture_snr and target_snr
        ssnrmean = 10 * np.log10(np.mean(sub_segnsr + 1e-10))
        # seg SNR max
        ssnrmax = 10 * np.log10(max(sub_segnsr + 1e-10))
        # seg SNR 80% percentile
        ssnrpk80 = 10 * np.log10(np.percentile(sub_segnsr + 1e-10, 80, interpolation='midpoint'))

        # Truth and neural-net prediction data, feature_frames x num_classes
        sub_truth = truth[o_frame_begin:o_frame_end, ]

        if predict is not None:
            sub_predict = predict[o_frame_begin:o_frame_end]
            # don't do this, expect genft to do the right thing w/truth, must be mutex
            # special case for false alarm analysis, set truth to zeros
            # if float(target_snr) < -96:
            #     truth = np.zeros(truth.shape)

            # ACC TPR PPV TNR FPR HITFA F1 MCC NT PT for each class, nclass x 10
            # cm is [TN FP; FN TP]
            (mcm, metrics, cm, cmn, rmse) = one_hot(sub_truth, sub_predict, cmode)
            # mcm2, metrics2, cm2, cmn2, rmse2 = classmetrics(sub_truth, sub_predict, cmode)

            # Save all stats, metrics and TN FP FN TP, but only includes for truth_index class.
            mstat[ni, ti, target_snr_index, augi,] = [mi, target_snr, ssnrmean, ssnrmax, ssnrpk80, target_gain,
                                                      *metrics[truth_index,],
                                                      mcm[truth_index, 0, 0], mcm[truth_index, 0, 1],
                                                      mcm[truth_index, 1, 0],
                                                      mcm[truth_index, 1, 1],
                                                      rmse[truth_index]]
            # metricsall[ni, ti, target_snr_index, augi, ] = metrics
            mcmall[ni, ti, target_snr_index, augi,] = mcm
            cmall[ni, ti, target_snr_index, augi,] = cm
            # cmnall[ni, ti, target_snr_index, augi, ] = cmn

            # Increment augmentation type
            augi += 1
            if augi >= num_augmentation_types:
                augi = 0

        if mi + 1 in plot_number:
            tname = Path(target_files[ti]).stem
            nname = Path(noise_files[ni]).stem
            logger.debug(
                f'{mi:{miwidth}}: Mix spec: SNR: {float(target_snr):4.1f}, target gain: {target_gain:+4.1f}, file: {tname:>{tfwidth}}, {nname:>{nfwidth}}')
            logger.debug(
                f'{mi:{miwidth}}: Measured: SNR: {mixture_snr:4.1f}, target level: {target_level:4.1f}, SegSNR mean, max, peak80%: {ssnrmean:4.1f}, {ssnrmax:4.1f}, {ssnrpk80:4.1f}')

            # Expand number of frames to number of samples (for plotting together with waveforms)
            segsnrex = np.reshape(np.tile(10 * np.log10(sub_segnsr + np.finfo(float).eps), [samples_per_transform, 1]),
                                  len(sub_segnsr) * samples_per_transform,
                                  order='F')
            sub_feature = feature[o_frame_begin:o_frame_end, ]

            # Truth extension to #samples in this mixture, i.e. MixSubSamples x num_classes
            truthex = np.zeros((len(sub_segnsr) * samples_per_transform, num_classes))
            for ntri in range(num_classes):
                truthex[:, ntri] = np.reshape(np.tile(sub_truth[:, ntri], [samples_per_feature, 1]),
                                              len(sub_segnsr) * samples_per_transform, order='F')

            # vector of indices where truth is present
            truth_active = []
            for i, x in enumerate(np.any(sub_truth, axis=0)):
                if x:
                    truth_active.append(i)

            if truth_is_mutex:
                # remove mutex (end), but not when class is always active
                if len(truth_active) > 1:
                    truth_active = np.delete(truth_active, -1)

            if len(truth_active) != 0:
                logger.debug(
                    f'{mi:{miwidth}}: Truth present in {len(truth_active)} of {num_classes} classes. First active category index: {truth_active[0]}')
            else:
                logger.debug(f'{mi:{miwidth}}: Truth not present in {num_classes} classes.')

            fig, (ax1, ax2, ax3) = plt.subplots(3, 1)

            secu = np.arange(len(sub_segnsr) * samples_per_transform, dtype=np.single) / sonusai.mixture.sample_rate
            plots = []
            ymin = [0]

            if mixture is not None:
                mix_plot, = ax1.plot(secu, sub_mixture, 'r', label='Mix')
                plots.append(mix_plot)
                ymin.append(min(sub_mixture))
            if target is not None:
                target_plot, = ax1.plot(secu, sub_target, 'b', label='Target')
                plots.append(target_plot)
                ymin.append(min(sub_target))

            ax1.set_ylim([min(ymin), 1.05])

            if len(truth_active) > 0:
                if len(truth_active) == 1:
                    truthn_plot, = ax1.plot(secu, truthex[:, truth_active[0]], 'g-',
                                            label=f'NN Truth{truth_active[0]}')
                    plots.append(truthn_plot)
                    if truth_is_mutex:
                        trutho_plot, = ax1.plot(secu, truthex[:, -1], 'm-.',
                                                label=f'NN Truth{num_classes} (Other)')
                        plots.append(trutho_plot)
                else:
                    trutha_plot, = ax1.plot(secu, truthex, 'g-', label='NN Truth All')
                    plots.append(trutha_plot)

            if sub_predict is not None:
                nnoex = np.reshape(np.tile(sub_predict, [samples_per_feature, 1]),
                                   len(sub_segnsr) * samples_per_transform)
                nno_plot, = ax1.plot(secu, nnoex, 'k', label='NN Predict All')
                plots.append(nno_plot)

            ax1.legend(handles=plots, bbox_to_anchor=(1.05, 1.0), loc='upper left')

            ax2.plot(secu, segsnrex)
            ax2.set_ylim([-51, max(min(max(segsnrex), 100), -10)])
            ax2.title.set_text(
                f'SegSNR mean={ssnrmean:4.1f} dB, max={ssnrmax:4.1f} dB, Peak80%={ssnrpk80:4.1f} dB')

            if sub_feature.shape[1] < 5:
                tmp = np.reshape(sub_feature, (sub_feature.shape[0], sub_feature.shape[1] * sub_feature.shape[2]))
                titlestr = f'All {sub_feature.shape[0]} features of size {sub_feature.shape[2]}x{sub_feature.shape[1]}'
            else:
                if len(truth_active) != 0:
                    fti = np.nonzero(sub_truth[:, truth_active[0]])[0][0]
                else:
                    fti = 0
                tmp = np.squeeze(sub_feature[fti,])
                titlestr = f'Example feature, frame {fti} of size {sub_feature.shape[2]}x{sub_feature.shape[1]}'
            ax3.imshow(np.transpose(tmp), aspect='auto')
            ax3.invert_yaxis()
            ax3.title.set_text(titlestr)

            fig.suptitle(f'{mi + 1} of {num_mixtures}')
            fig.tight_layout()
            pdf.savefig(fig)
            plt.close(fig)

    if plot_number:
        pdf.close()

    if predict is not None:
        save_eval_report(output_dir=output_dir,
                         noise_files=noise_files,
                         target_files=target_files,
                         snrs=snrs,
                         fnameidx=fnameidx,
                         binary_detection_threshold=binary_detection_threshold,
                         np_snrs=np_snrs,
                         truth_is_mutex=truth_is_mutex,
                         mstat=mstat,
                         mcmall=mcmall,
                         cmall=cmall)


def median_filter(x: np.ndarray, n: int) -> np.ndarray:
    assert x.ndim == 1, 'Input must be one-dimensional.'

    n2 = (n - 1) // 2
    if n % 2 == 0:
        n2 += 1

    pad_x = np.zeros(len(x) + n, dtype=x.dtype)
    pad_x[n2:n2 + len(x)] = x

    y = np.zeros((len(x), n), dtype=x.dtype)
    for k in range(len(x)):
        y[k,] = pad_x[k:k + n]

    return np.median(y, axis=1)


def segsnr_filter(segsnr: np.ndarray) -> np.ndarray:
    assert segsnr.ndim == 1, 'Input must be one-dimensional.'

    inf_indices = [i for i, val in enumerate(segsnr) if np.isinf(val)]

    if inf_indices:
        scale = np.arange(16) + 1
        for inf_index in inf_indices:
            # avoid underflow at beginning
            if inf_index > 15:
                i = range(inf_index - 1, inf_index - 17, -1)
                segsnr[inf_index] = np.sum(np.multiply(segsnr[i], scale)) / 136
            else:
                i = range(inf_index + 1)
                segsnr[inf_index] = np.median(segsnr[i])
        segsnr = median_filter(segsnr, 32)
    return segsnr


def save_eval_report(output_dir: str,
                     noise_files: list,
                     target_files: list,
                     snrs: dict,
                     fnameidx: np.ndarray,
                     binary_detection_threshold: float,
                     np_snrs: list,
                     truth_is_mutex: bool,
                     mstat: np.ndarray,
                     mcmall: np.ndarray,
                     cmall: np.ndarray,
                     nfacrit: float = 0.05):
    outname = 'evaluate'
    num_noise_files = len(noise_files)
    num_target_files = len(target_files)
    num_snrs = len(snrs)
    num_augmentation_types = mstat.shape[3]
    NCLASS = mcmall.shape[4]  # mcm will = 1 for binary (always be correct)
    NMIX = num_noise_files * num_target_files * num_snrs * num_augmentation_types
    logger.info(
        f'Printing reports in {output_dir} for {NMIX} mixtures, {num_noise_files} noise files, {num_target_files} target files, {num_snrs} SNRs, {num_augmentation_types} augmentations')

    # Target all SNR list index, ordered from highest to lowest SNR
    tasnridx = np.flip(np.argsort(mstat[0, 0, :, 0, 1]))
    # Detect special case SNR
    if len(np_snrs) == 0:
        np_snrs = len(snrs) - 1
        logger.warning(
            f'Did not find low SNR (<-96 dB) case, FA analysis uses lowest SNR of {float(snrs[np_snrs]):3.0f}\n')
    else:
        np_snrs = np_snrs[0]  # In case a list is returned, use 0, TBD list support

    FPSNR = float(snrs[np_snrs])
    # Create list of non-special case SNRs, for some metrics
    if FPSNR < -96:
        sridx = np.arange(0, num_snrs)
        sridx = np.delete(sridx, np_snrs)  # Remove special case SNR
        srsnrlist = list(map(snrs.__getitem__, sridx))  # new list wo fpsnr case
    else:
        FPSNR = 99  # Mark as unused with value 99
        sridx = tasnridx
        srsnrlist = snrs

    snrfloat = np.zeros((num_snrs, 1))
    for i, x in enumerate(snrs):
        snrfloat[i] = float(x)

    if NCLASS == 1:
        bmode = True
    else:
        bmode = False  # TBD might need to detect NCLASS == 2 and truth_is_mutex

    # Total metrics over all data
    metsa, mcmsa, cmsumsa = calculate_metrics(cmall, bmode)  # metric summary all
    metsa_avg = averages(metsa)  # 3 rows: [PPV, TPR, F1, FPR, ACC, TPSUM]
    metsa_avgwo = averages(metsa[0:-1, ])

    # Total metrics over all non-special case SNR
    metsn, mcmsn, cmsumsn = calculate_metrics(cmall[:, :, sridx, ], bmode)
    metsn_avg = averages(metsn)  # 3x6
    metsn_avgwo = averages(metsn[0:-1, ])  # without other (last) class

    # Metric summary by noise and snr per class: 
    # fields: [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]
    metsnr = np.zeros((num_snrs, NCLASS, 12))  # per SNR: nsnr x class x 12
    metsnr_avg = np.zeros((num_snrs, 3, 6))  # per SNR avg over classes: nsnr x 3 x 6
    metsnr_avgwo = np.zeros((num_snrs, 3, 6))  # without other class
    metnsnr = np.zeros((num_noise_files, num_snrs, NCLASS, 12))  # Per Noise & SNR: nnf x nsnr x nclass x 12
    metnsnr_avg = np.zeros((num_noise_files, num_snrs, 3, 6))
    metnsnr_avgwo = np.zeros((num_noise_files, num_snrs, 3, 6))
    for target_snr_index in range(num_snrs):
        metsnr[target_snr_index,], _, _ = calculate_metrics(cmall[:, :, target_snr_index, ])
        metsnr_avg[target_snr_index,] = averages(metsnr[target_snr_index,])
        metsnr_avgwo[target_snr_index,] = averages(metsnr[target_snr_index, 0:-1, ])  # without other class = nclass
        for ni in range(num_noise_files):
            metnsnr[ni, target_snr_index,], _, _ = calculate_metrics(cmall[ni, :, target_snr_index, ])
            metnsnr_avg[ni, target_snr_index,] = averages(metnsnr[ni, target_snr_index,])
            metnsnr_avgwo[ni, target_snr_index,] = averages(metnsnr[ni, target_snr_index, 0:-1, ])

    # mcm: [TN FP; FN TP]
    # Fields in mstat, TargetSNR, TargetGain: 
    FTSNR = 1
    FTGAIN = 5
    # Fields in metrics: 
    FACC = 0
    FF1 = 6
    FTPR = 1
    FFA = 4
    FHITFA = 5

    # FA per noise-file @ lowest SNR, averages over classes:
    fanf_macwo = metnsnr_avgwo[:, np_snrs, 0, 3]  # macro-average wo other class NNFx1
    fanf_micwo = metnsnr_avgwo[:, np_snrs, 1, 3]  # micro-average wo other class NNFx1
    fanf_mac = metnsnr_avg[:, np_snrs, 0, 3]  # macro-average NNFx1
    fanf_mic = metnsnr_avg[:, np_snrs, 1, 3]  # micro-average NNFx1
    nfafail = [i for i, x in enumerate(fanf_mic) if x > nfacrit]

    # Create summary report
    summary_name = f'{output_dir}/nn_perfsummary.txt'
    with open(file=summary_name, mode='w') as f:
        # RMSEmean = np.squeeze(np.mean(np.reshape(mstat[:, :, :, :, 20], (num_noise_files, num_target_files * num_augmentation_types * num_snrs, 1)), 1))

        if bmode:  # -------------- Binary Classification ------------------
            f.write('\n')
            f.write(f'--- Aaware Binary NN Performance Analysis: {outname}\n')
            f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
            f.write(
                f'{num_target_files} target files and {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
            f.write(f'Binary classification threshold: {binary_detection_threshold:4.2f}.\n')
            f.write('\n')
            # FA Summary, note: for binary *_avg are accurate, although no avg needed (but avgwo is not)
            f.write(f'--- False Alarm Performance over {num_noise_files} noise files  ---\n')
            f.write(f'FA over all mixtures:             {metsa_avg[1, 3] * 100:6.5f}%\n')
            f.write(f'Max FA over SNRs and Noise files: {np.max(metnsnr_avg[:, :, 1, 3]) * 100:6.5f}%\n')
            f.write(f'FA of lowest SNR {FPSNR:2.0f} dB:          {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%\n')
            if len(nfafail) > 0:
                f.write(f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
            else:
                f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')

            f.write(generate_snr_summary(mstat, metsnr, snrfloat, tasnridx))

            f.write('\n\n')
            # Failing Noise File List
            if len(nfafail) > 0:
                f.write(
                    f'--- List of {len(nfafail)} failing noise files with FA rate > {nfacrit * 100:3.1f}%:\n')
                for jj in range(len(nfafail)):
                    f.write(f'{noise_files[nfafail[jj]]:40s}: {100 * fanf_mic[nfafail[jj]]:3.1f}%\n')

            f.write('\n')

        else:  # ----------------- Multiclass Classification -----------------------------
            f.write('\n')
            f.write(f'--- Aaware Multiclass NN Performance Analysis: {outname}\n')
            f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
            f.write(
                f'{num_target_files} target files and {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
            f.write(f'Number of classes: {NCLASS}.\n')
            f.write(f'Mutex mode (single-label): {truth_is_mutex}\n')
            f.write(f'Classification decision threshold: {binary_detection_threshold:4.2f}.\n')
            f.write('Confusion Matrix over all mixtures:\n')
            f.write(generate_snr_summary(mstat, metsnr, snrfloat, tasnridx))
            f.write('\n')
            # Metrics over all mixtures:
            f.write('Confusion Matrix over all mixtures:\n')
            f.write(format_confusion_matrix(cmsumsa))
            f.write('Metric summary, all mixtures:\n')
            f.write(generate_summary(metsa))
            f.write('\n')
            # Metrics excluding special case SNR
            if FPSNR < -96:
                f.write(
                    f'Metrics excluding special case false-positive SNR (FPSNR) at {FPSNR:2.0f} dB\n')
                f.write('Confusion Matrix for all mixtures except FPSNR:\n')
                f.write(format_confusion_matrix(cmsumsn))

            # FA Summary
            f.write(
                f'False alarm over all mixtures, class micro-avg for all, non-other: {metsa_avg[1, 3] * 100:6.5f}%, {metsa_avgwo[1, 3] * 100:6.5f}%\n')
            f.write(
                f'False alarm summary for lowest SNR at {FPSNR:2.0f} dB, class micro-avg over all, non-other: {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%, {metsnr_avgwo[np_snrs, 1, 3] * 100:6.5f}%\n')
            if len(nfafail) > 0:
                f.write(
                    f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)} (see noise-breakdown report)\n')
            else:
                f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
            # Print false alarm summary, use micro-avg which tends more pessimistic excluding other class if in mutex mode

            f.write('\n\n')
            f.write('-------- Per SNR Metrics ------------ \n')
            f.write('\n')
            for target_snr_index in range(num_snrs):
                si = tasnridx[target_snr_index]  # sorted list
                f.write('\n')
                f.write(f'--- Confusion Matrix for SNR {float(snrs[si]):3.1f} dB:\n')
                (met, _, cmsum) = calculate_metrics(cmall[:, :, si, ])
                f.write(format_confusion_matrix(cmsum))
                f.write('Metric summary:\n')
                f.write(generate_summary(metsnr[si,]))
                f.write('\n\n')

    # --------------- Create noise breakdown report --------------------------------
    noise_breakdown_name = f'{output_dir}/noise_breakdown.txt'
    with open(file=noise_breakdown_name, mode='w') as f:
        # Calc #frames per noise file, all speech,augm cases at lowSNR case
        f.write('\n')
        f.write(f'--- Aaware sonusai() NN False Alarm Analysis: {outname}\n')
        f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
        f.write(
            f'{num_target_files} target files, {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
        f.write(f'Classification decision threshold: {binary_detection_threshold:4.2f}.\n')
        f.write(f'Number of classes: {NCLASS}.\n')
        f.write(f'Mutex mode (single-label): {truth_is_mutex}\n')
        f.write('\n')
        f.write(f'--- NN False Alarm Summary over {num_noise_files} noise files ---\n')
        # FA Summary
        # Print false alarm summary, use micro-avg which tends more pessimistic excluding other class if in mutex mode
        f.write(
            f'False alarm over all mixtures, class micro-avg for all, non-other: {metsa_avg[1, 3] * 100:6.5f}%,{metsa_avgwo[1, 3] * 100:6.5f}%\n')
        f.write(
            f'False alarm summary for lowest SNR at {FPSNR:2.0f} dB, class micro-avg over all, non-other: {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%, {metsnr_avgwo[np_snrs, 1, 3] * 100:6.5f}%\n')

        if len(nfafail) > 0:
            f.write(f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
        else:
            f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
        # FA per SNR
        f.write('False alarm per SNR micro-avg (%):\n')
        f.write(f'{np.transpose(snrfloat[tasnridx].astype(int))}\n')
        f.write(f'{np.round(metsnr_avg[tasnridx, 1, 3] * 100, 2)}\n')
        # FA per SNR and Class:
        # f.write(f'False alarm per SNR and Class (%):\n {np.round(metsnr[:,:,FFA]*100,2)}\n')

        f.write('\n')
        NFLOWSNR = int(
            np.sum((metnsnr[:, np_snrs, 0, 8] + metnsnr[:, np_snrs, 0, 9])) / num_noise_files)  # Nframes = NT + PT
        f.write(
            f'--- FA perf. per noise file, over {num_target_files * num_augmentation_types} mixtures, {NFLOWSNR} frames per SNR:\n')

        nfileidx = fnameidx[:, 0, 0, 0, 0]
        for ni in range(num_noise_files):
            # sortni = issnrl[ni]   # TBD figure out how to sort by FA perf
            # FA per SNR
            tmpnfname = splitext(basename(noise_files[nfileidx[ni]]))[0]
            f.write('\n')
            f.write(f'--- NN Perf. for noise file {ni + 1} of {num_noise_files}: {tmpnfname} ---\n')
            f.write('False alarm per SNR micro-avg (%):\n')
            f.write(f'{np.transpose(snrfloat[tasnridx].astype(int))}\n')
            f.write(f'{np.round(metnsnr_avg[ni, tasnridx, 1, 3] * 100, 2)}\n')
            f.write('Confusion Matrix:\n')
            (_, _, cmsum) = calculate_metrics(cmall[ni,])
            f.write(format_confusion_matrix(cmsum))
            f.write(f'Noise file full path: {noise_files[nfileidx[ni]]}\n')


def get_level(audio: np.ndarray) -> float:
    if audio is not None:
        # level in dB
        return 20 * np.log10(np.sqrt(np.mean(np.square(audio))) + np.finfo(float).eps)
    return float('nan')


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        feature_name = args['--feature']

        if args['--mixture']:
            mixture_name = args['--mixture']
        else:
            mixture_name = ''

        if args['--predict']:
            predict_name = args['--predict']
        else:
            predict_name = ''

        binary_detection_threshold = float(args['--bthr'])

        plot_number = int(args['--plotnum'])

        # create output directory
        output_dir = f'evaluate-{datetime.now():%Y%m%d-%H%M%S}'
        try:
            mkdir(output_dir)
        except OSError as error:
            logger.error(f'Could not create directory, {output_dir}: {error}')
            raise SystemExit(1)

        log_name = output_dir + '/evaluate.log'
        create_file_handler(log_name)

        mixdb, feature, truth, segsnr = read_feature_data(feature_name)
        mixture, target, noise = read_mixture_data(mixture_name)
        predict = read_predict_data(predict_name, feature.shape[0])

        evaluate(mixdb=mixdb,
                 feature=feature,
                 truth=truth,
                 segsnr=segsnr,
                 output_dir=output_dir,
                 mixture=mixture,
                 target=target,
                 noise=noise,
                 predict=predict,
                 binary_detection_threshold=binary_detection_threshold,
                 plot_number=plot_number,
                 verbose=args['--verbose'])

        logger.info(f'Wrote results to {output_dir}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



from typing import Union


def create_onnx_from_keras(model,
                           name: str = 'model',
                           file_pfx: Union[None, str] = None,
                           is_flattened: bool = True,
                           has_timestep: bool = True,
                           has_channel: bool = False,
                           is_mutex: bool = True,
                           feature: str = ''):
    # Create onnx model from keras model and write onnx model file if file_pfx provided.
    #   model           keras model
    #   name            model name
    #   file_pfx        filename prefix to save onnx model (do not save if empty)
    #   is_flattened    model feature data is flattened
    #   has_timestep    model has timestep dimension
    #   has_channel     model has channel dimension
    #   is_mutex        model label output is mutually exclusive
    #   feature         model feature type
    # import tf2onnx
    import keras2onnx
    from tensorflow.keras.layers import GRU

    msg = ''

    # convert to onnx model
    # spec = (tf.TensorSpec((None, 224, 224, 3), tf.float32, name="input"),)
    # onnx_model, _ = tf2onnx.convert.from_keras(model)
    onnx_model = keras2onnx.convert_keras(model, name, debug_mode=True, target_opset=13)

    # Replace stateful GRUs with custom layers
    stateful_gru_names = []
    for i in range(len(model.layers)):
        layer = model.layers[i]
        if isinstance(layer, GRU):
            if layer.stateful:
                stateful_gru_names.append(layer.name)

    for i in range(len(onnx_model.graph.node)):
        node = onnx_model.graph.node[i]
        if node.name in stateful_gru_names:
            node.op_type = 'SGRU'

    # Add metadata to model
    f_flag = onnx_model.metadata_props.add()
    f_flag.key = 'is_flattened'
    f_flag.value = str(is_flattened)

    t_flag = onnx_model.metadata_props.add()
    t_flag.key = 'has_timestep'
    t_flag.value = str(has_timestep)

    c_flag = onnx_model.metadata_props.add()
    c_flag.key = 'has_channel'
    c_flag.value = str(has_channel)

    m_flag = onnx_model.metadata_props.add()
    m_flag.key = 'is_mutex'
    m_flag.value = str(is_mutex)

    feature_flag = onnx_model.metadata_props.add()
    feature_flag.key = 'feature'
    feature_flag.value = str(feature)

    # save the model in ONNX format
    if file_pfx is not None:
        file_name = file_pfx + '.onnx'
        msg = f'onnx model saved to {file_name}'
        keras2onnx.save_model(onnx_model, file_name)

    return onnx_model, msg



# SonusAI general utilities

from sonusai.utils.expandvars import expandvars
from sonusai.utils.grouper import grouper
from sonusai.utils.human_readable_size import human_readable_size
from sonusai.utils.int16_to_float import int16_to_float
from sonusai.utils.p_tqdm import p_map
from sonusai.utils.p_tqdm import p_umap
from sonusai.utils.read_feature_data import read_feature_data
from sonusai.utils.read_mixture_data import read_mixture_data
from sonusai.utils.read_predict_data import read_predict_data
from sonusai.utils.reshape_inputs import reshape_inputs
from sonusai.utils.seconds_to_hms import seconds_to_hms
from sonusai.utils.to_numpy_array import to_numpy_array
from sonusai.utils.trim_docstring import trim_docstring
from sonusai.utils.yes_or_no import yes_or_no



import numpy as np


def estimate_audio_length(audio_in: np.ndarray,
                          augmentation: dict,
                          length_common_denominator: int) -> int:
    length = len(audio_in)

    if 'tempo' in augmentation:
        factor = augmentation['tempo']
        length = int(length / factor)

    if length % length_common_denominator:
        length += length_common_denominator - int(length % length_common_denominator)

    return length



from typing import List

import numpy as np

from sonusai.metrics import averages


def generate_snr_summary(mstat: np.ndarray, smetrics: np.ndarray, snrfloat: List[float], tasnridx: List[int]) -> str:
    # Generate summary for each SNR
    # mstat:    config params like target snr, levels, etc. [NNF, NTF, NSNR, NAUG, 23]
    # smetrics: metric data summed by SNR, must be shape [NSNR, num_classes, 12]
    # snrfloat: snr values for each NSNR dim in mstat & metsnr
    # tasnridx: index into snrfloat sorted (i.e. for highest to lowest snr)
    #
    # mstat fields:
    # [mi, tsnr, ssnrmean, ssnrmax, ssnrpk80, tgain, metrics[tridx,12], TN, FN, FP, TP, rmse[tridx]
    # metsnr fields: ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]

    assert mstat.ndim == 5, 'mstat must be 5-dimensional.'

    NNF, NTF, NSNR, NAUG, _ = mstat.shape
    NMIX = NNF * NTF * NSNR * NAUG
    num_classes = smetrics.shape[1]

    result = f'--- NN Performance over {NMIX / NSNR} mixtures per Global SNR ---\n'
    result += '| SNR |  PPV% |  TPR% |  F1%  |  FPR% |  ACC% | SgSNRavg | SgSNR80p |\n'
    for si in range(NSNR):
        snri = tasnridx[si]
        tmpif = np.isfinite(mstat[:, :, snri, :, 2])
        mssnravg = np.mean(mstat[:, :, snri, :, 2][tmpif])  # mean segmsnr, unweighted avg, ignoring nans
        tmpif = np.isfinite(mstat[:, :, snri, :, 4])
        segsnr80pc = np.mean(mstat[:, :, snri, :, 4][tmpif])  # segmsnr80pc, unweighted avg, ignoring nans
        metavg = averages(smetrics[snri,])  # multiclass uses class averages
        # metavg fields: [PPV, TPR, F1, FPR, ACC, TPSUM]
        result += f'| {max(min(round(mstat[0, 0, snri, 0, 1]), 99), -99):+3.0f} '
        result += f'| {metavg[0, 0] * 100:5.1f} '
        result += f'| {metavg[0, 1] * 100:5.1f} '
        result += f'| {metavg[0, 2] * 100:5.1f} '
        result += f'| {metavg[1, 3] * 100:5.1f} '
        result += f'| {metavg[1, 4] * 100:5.1f} '
        result += f'|   {max(min(round(mssnravg), 99), -99):+3.0f}    '
        result += f'|   {max(min(round(segsnr80pc), 99), -99):+3.0f}    '
        result += f'|\n'

    if num_classes > 1:
        result += f'PPV,TPR,F1 are macro-avg, FPR, ACC are micro-avg over {num_classes:>6} classes.\n'

    return result
